﻿namespace FaceDetection
{
  using System;
  using System.Collections.Generic;
  using System.Drawing;
  using System.Drawing.Drawing2D;
  using System.Drawing.Imaging;
  using System.IO;
  using System.Linq;
  using System.Runtime.InteropServices;
  using Ionic.BZip2;
  
  internal static class Converter
  {
    private const string OutputPath = @"Images\Positives";

    public static void Convert(string inputPath)
    {
      Directory.CreateDirectory(OutputPath);
      var truthsPath = Path.Combine(inputPath, "data", "ground_truths", "name_value");
      foreach (var dirpath in Directory.EnumerateDirectories(truthsPath))
      {
        foreach (var filepath in Directory.EnumerateFiles(dirpath))
        {
          if (Path.GetFileName(filepath).Contains('_'))
          {
            ConvertFile(filepath, inputPath);
          }
        }
      }
    }

    private static Dictionary<string, string> ReadTruth(string path)
    {
      var truths = new Dictionary<string, string>();
      using (var reader = new StreamReader(path))
      {
        while (true)
        {
          var line = reader.ReadLine();
          if (String.IsNullOrEmpty(line))
          {
            break;
          }
          var parts = line.Split('=');
          if (parts.Length == 2)
          {
            truths[parts[0]] = parts[1];
          }
        }
      }
      return truths;
    }

    private static void ConvertFile(string filepath, string inputPath)
    {
      const int ppmWidth = 512;
      const int ppmHeight = 768;

      var truth = ReadTruth(filepath);
      if (!truth.ContainsKey("left_eye_coordinates") || truth["glasses"] == "Yes" || truth["beard"] == "Yes" || truth["mustache"] == "Yes")
      {
        return;
      }

      byte[] bytes = new byte[ppmWidth * ppmHeight * 4];
      var ppmPath = Path.Combine(inputPath, truth["relative"]);
      using (var compressed = new FileStream(ppmPath, FileMode.Open, FileAccess.Read))
      using (var stream = new BZip2InputStream(compressed))
      {
        for (int i = 0; i < 3; ++i)
        {
          while (stream.ReadByte() != '\n')
          {
            continue;
          }
        }
        
        for (int i = 0; i < bytes.Length; i += 4)
        {
          bytes[i + 2] = (byte)stream.ReadByte();
          bytes[i + 1] = (byte)stream.ReadByte();
          bytes[i + 0] = (byte)stream.ReadByte();
        }
      }

      using (var original = new Bitmap(ppmWidth, ppmHeight, PixelFormat.Format32bppRgb))
      {
        var data = original.LockBits(new Rectangle(0, 0, ppmWidth, ppmHeight), ImageLockMode.WriteOnly, original.PixelFormat);
        Marshal.Copy(bytes, 0, data.Scan0, bytes.Length);
        original.UnlockBits(data);

        var left = truth["left_eye_coordinates"].Split(' ').Select(p => Int32.Parse(p)).ToArray();
        var right = truth["right_eye_coordinates"].Split(' ').Select(p => Int32.Parse(p)).ToArray();

        int padding = (left[0] - right[0]) / 4;
        int length = padding * 6;
        float angle = (float)(Math.Atan2(left[1] - right[1], left[0] - right[0]) / Math.PI * 180);

        using (var corrected = new Bitmap(length, length, PixelFormat.Format24bppRgb))
        {
          using (var gfx = Graphics.FromImage(corrected))
          {
            gfx.InterpolationMode = InterpolationMode.HighQualityBicubic;
            gfx.CompositingQuality = CompositingQuality.HighQuality;
            gfx.SmoothingMode = SmoothingMode.HighQuality;
            
            gfx.TranslateTransform(-right[0], -right[1], MatrixOrder.Append);
            gfx.RotateTransform(-angle, MatrixOrder.Append);
            gfx.TranslateTransform(padding, padding, MatrixOrder.Append);
            gfx.DrawImage(original, Point.Empty);
          }
          
          var path = Path.GetFileNameWithoutExtension(truth["relative"]);
          var encoder = ImageCodecInfo.GetImageEncoders().First(e => e.FormatID == ImageFormat.Jpeg.Guid);
          var parameters = new EncoderParameters();
          parameters.Param[0] = new EncoderParameter(Encoder.Quality, 95L);
          corrected.Save(Path.Combine(OutputPath, Path.ChangeExtension(path, "jpg")), encoder, parameters);
        }
      }
    }
  }
}
