﻿namespace FaceDetection
{
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using SharpNeat.Core;
  using SharpNeat.Phenomes;
  using SharpNeat.Utility;
  using System.IO;
  
  internal sealed class Evaluator : IPhenomeEvaluator<IBlackBox>
  {
    private const double PositiveOutput = 1;
    private const double NegativeOutput = 0;

    private const double PositiveError = 0.1;
    private const double NegativeError = 1 - PositiveError;

    private readonly double factor;
    private readonly FastRandom random;
    private readonly List<byte[]> allPositives;
    private readonly List<byte[]> allNegatives;
    private readonly byte[] averagePositive;

    private readonly HashSet<int> positives;
    private readonly HashSet<int> negatives;

    public Evaluator(Loader loader, double factor, bool initSets = true)
    {
      this.factor = factor;
      random = new FastRandom();
      allPositives = loader.GetPositives();
      allNegatives = loader.GetNegatives();

      var total = new int[allPositives[0].Length];
      foreach (var positive in allPositives)
      {
        for (int i = 0; i < total.Length; ++i)
        {
          total[i] += positive[i];
        }
      }

      averagePositive = new byte[total.Length];
      for (int i = 0; i < total.Length; ++i)
      {
        averagePositive[i] = (byte)(total[i] / allPositives.Count);
      }

      if (initSets)
      {
        positives = new HashSet<int>();
        while (positives.Count != 500)
        {
          positives.Add(random.Next(allPositives.Count));
        }

        negatives = new HashSet<int>();
        while (negatives.Count != 1500)
        {
          negatives.Add(random.Next(allNegatives.Count));
        }
      }
    }

    public ulong EvaluationCount { get; set; }

    public FitnessInfo Evaluate(IBlackBox box)
    {
      ++EvaluationCount;

      double positivesFitness = positives.Count;
      int positivesCorrect = 0;
      foreach (var index in positives)
      {
        double error = CalculateError(box, allPositives[index], PositiveOutput);
        positivesFitness -= error < 0.5 ? error * error : 1;
        if (error < PositiveError)
        {
          ++positivesCorrect;
        }
      }

      double negativesFitness = negatives.Count;
      int negativesCorrect = 0;
      foreach (var index in negatives)
      {
        double error = CalculateError(box, allNegatives[index], NegativeOutput);
        negativesFitness -= error < 0.5 ? error * error : 1;
        if (error < NegativeError)
        {
          ++negativesCorrect;
        }
      }

      double fitness = factor * positivesFitness / positives.Count + (1 - factor) * negativesFitness / negatives.Count;
      return new FitnessInfo(fitness * 100, null);
    }

    public double CalculateError(IBlackBox box, byte[] inputs, double expected)
    {
      box.ResetState();
      for (int i = 0; i < inputs.Length; ++i)
      {
        box.InputSignalArray[i] = Math.Abs(inputs[i] - averagePositive[i]) / 255.0;
      }

      box.Activate();
      return Math.Abs(box.OutputSignalArray[0] - expected);
    }

    public void Reset()
    {
    }

    public bool StopConditionSatisfied { get; set; }

    public void ExpandSets(IList<IBlackBox> boxes)
    {
      ExpandSet(positives, allPositives, 10, boxes, PositiveOutput);
      ExpandSet(negatives, allNegatives, 30, boxes, NegativeOutput);
    }

    private void ExpandSet(HashSet<int> current, List<byte[]> all, int count, IList<IBlackBox> boxes, double output)
    {
      var toAdd = Enumerable.Range(0, all.Count)
        .Where(i => !current.Contains(i))
        .Select(i => new
        {
          Index = i,
          Error = boxes.AsParallel().Sum(box => CalculateError(box, all[i], output))
        })
        .Where(i => i.Error > 0.5)
        .OrderByDescending(i => i.Error)
        .Select(i => i.Index)
        .Take(count);

      foreach (var index in toAdd)
      {
        current.Add(index);
      }
    }

    public string FullyValidate(IBlackBox box)
    {
      var positivesCorrect = allPositives.Where(p => CalculateError(box, p, PositiveOutput) < PositiveError).Count();
      var negativesCorrect = allNegatives.Where(n => CalculateError(box, n, NegativeOutput) < NegativeError).Count();
      return String.Format("{0:0.00000}\t{1:0.00000}",
        (double)positivesCorrect / allPositives.Count * 100, (double)negativesCorrect / allNegatives.Count * 100);
    }

    public IEnumerable<string> CreateScoutingMatrix(IList<IBlackBox> boxes)
    {
      for (int i = 0; i < allPositives.Count; ++i)
      {
        yield return CreateScoutingRow(allPositives[i], PositiveOutput, PositiveError, boxes);
      }
      for (int i = 0; i < allNegatives.Count; ++i)
      {
        yield return CreateScoutingRow(allNegatives[i], NegativeOutput, NegativeError, boxes);
      }
    }

    private string CreateScoutingRow(byte[] inputs, double output, double error, IList<IBlackBox> boxes)
    {
      return String.Concat(boxes.AsParallel().AsOrdered()
        .Select(box => CalculateError(box, inputs, output) < error ? "0" : "1"));
    }
  }
}
